@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -226,6 +226,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
             if ( __validate_dbc( connection ))
             {
                 thread_protect( SQL_HANDLE_DBC, connection );
+                function_entry( connection );
                 dm_log_write( __FILE__, 
                         __LINE__, 
                         LOG_INFO, 
@@ -235,8 +236,8 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 __post_internal_error( &connection -> error,
                     ERROR_HY024, NULL,
                     connection -> environment -> requested_version );
-        
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
             else 
             {
@@ -264,6 +265,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 if ( __validate_dbc( connection ))
                 {
                     thread_protect( SQL_HANDLE_DBC, connection );
+                    function_entry( connection );
                     dm_log_write( __FILE__, 
                             __LINE__, 
                             LOG_INFO, 
@@ -274,7 +276,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                         ERROR_HY024, NULL,
                         connection -> environment -> requested_version );
             
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
                 else 
                 {
@@ -295,23 +297,25 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
             if ( __validate_dbc( connection ))
             {
                 thread_protect( SQL_HANDLE_DBC, connection );
+                function_entry( connection );
                 dm_log_write( __FILE__, 
                         __LINE__, 
                         LOG_INFO, 
                         LOG_INFO, 
-                        "Error: HY009" );
+                        "Error: HY024" );
         
                 __post_internal_error( &connection -> error,
-                    ERROR_HY009, NULL,
+                    ERROR_HY024, NULL,
                     connection -> environment -> requested_version );
         
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
             else 
             {
                 return SQL_INVALID_HANDLE;
             }
         }
+        return SQL_SUCCESS;
     }
 
     /*
@@ -365,7 +369,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 ERROR_08003, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
     else if ( connection -> state == STATE_C3 )
@@ -380,7 +384,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
             ERROR_HY010, NULL,
             connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
     else if ( connection -> state == STATE_C4 ||
             connection -> state == STATE_C5 )
@@ -397,7 +401,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 ERROR_08002, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
     else if ( connection -> state == STATE_C6 )
@@ -414,7 +418,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 ERROR_08002, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
         else if ( option == SQL_TXN_ISOLATION )
         {
@@ -428,7 +432,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 ERROR_S1011, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
@@ -449,7 +453,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 ERROR_HY024, NULL,
                 connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     /*
@@ -466,6 +470,16 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
         connection -> login_timeout_set = 1;
         connection -> login_timeout = value;
     }
+    else if ( option == SQL_ATTR_ACCESS_MODE )
+    {
+        connection -> access_mode = ( SQLLEN ) value;
+        connection -> access_mode_set = 1;
+    }
+    else if ( option == SQL_AUTOCOMMIT )
+    {
+        connection -> auto_commit = ( SQLINTEGER ) value;
+        connection -> auto_commit_set = 1;
+    }
 
     if ( option == SQL_ODBC_CURSORS )
     {
@@ -484,18 +498,48 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
             connection -> quite_mode = ( SQLLEN ) value;
             connection -> quite_mode_set = 1;
         }
+        else if ( option == SQL_ATTR_ACCESS_MODE )
+        {
+            connection -> access_mode = ( SQLLEN ) value;
+            connection -> access_mode_set = 1;
+        }
         else
         {
             /*
-             * save any unknown attributes untill connect
+             * save any unknown attributes until connect
              */
 
-            struct save_attr *sa = calloc( 1, sizeof( struct save_attr ));
+            struct save_attr sa, *sap;
+            
+            memset( &sa, 0, sizeof( sa ));
+            
+            sa.attr_type = option;
+            sa.intptr_attr = value;
+            
+            sap = connection -> save_attr;
+            
+            while ( sap )
+            {
+                if ( sap -> attr_type == option )
+                {
+                    free( sap -> str_attr );
+                    break;
+                }
+                sap = sap -> next;
+            }
 
-            sa -> attr_type = option;
-            sa -> int_attr = ( SQLINTEGER ) value;
-            sa -> next = connection -> save_attr;
-            connection -> save_attr = sa;
+            if ( sap )
+            {
+                *sap = sa;
+            }
+            else
+            {
+                sap = malloc( sizeof( struct save_attr ));
+                *sap = sa;
+
+                sap -> next = connection -> save_attr;
+                connection -> save_attr = sap;
+            }
         }
 
         if ( log_info.log_flag )
@@ -511,7 +555,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                     connection -> msg );
         }
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_SUCCESS );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_SUCCESS );
     }
     else
     {
@@ -538,7 +582,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                   case SQL_ATTR_TRACEFILE:
                   case SQL_ATTR_TRANSLATE_LIB:
                     string_length = SQL_NTS;
-                    ptr = (void *) ansi_to_unicode_alloc(( SQLCHAR * ) value, SQL_NTS, connection );
+                    ptr = (void *) ansi_to_unicode_alloc(( SQLCHAR * ) value, SQL_NTS, connection, NULL );
                     break;
 
                   default:
@@ -568,7 +612,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                         ERROR_IM001, NULL,
                         connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
         }
         else
@@ -599,7 +643,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                 ret = SQLSETCONNECTATTR( connection,
                         connection -> driver_dbc,
                         option,
-                        value,
+                        (SQLPOINTER)(intptr_t) value,
                         string_length );
             }
             else
@@ -614,7 +658,7 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
                         ERROR_IM001, NULL,
                         connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
         }
 
@@ -641,5 +685,5 @@ SQLRETURN SQLSetConnectOption( SQLHDBC connection_handle,
         connection -> bookmarks_on = (SQLUINTEGER) value;
     }
 
-    return function_return( SQL_HANDLE_DBC, connection, ret );
+    return function_return( SQL_HANDLE_DBC, connection, ret, DEFER_R3 );
 }
