@@ -5,10 +5,16 @@
 #define ODBCVER 0x0380
 #endif
 
+#ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#ifdef HAVE_PWD_H
 #include <pwd.h>
+#endif
 #include <ltdl.h>
+#ifdef HAVE_STRING_H
 #include <string.h>
+#endif
 
 #include <sqlext.h>                     /* THIS WILL BRING IN sql.h and
                                            sqltypes.h AS WELL AS PROVIDE
@@ -85,7 +91,7 @@ typedef struct cl_statement
     char                *sql_text;      /* text of current statement */
     char                **column_names; /* names of each column */
     SQLSMALLINT         *data_type;
-    SQLLEN              *column_size;
+    SQLULEN             *column_size;
     SQLSMALLINT         *decimal_digits;
     int                 driver_stmt_closed;
     int                 not_from_select;
@@ -99,7 +105,11 @@ typedef struct cl_statement
     int                 rowset_complete;
     FILE                *rowset_file;
     char                *rowset_buffer;
-    int                 buffer_length;
+#ifdef HAVE_FSEEKO
+    off_t               buffer_length;
+#else
+    long                buffer_length;
+#endif
     int                 column_count;
     int                 curr_rowset_start;
     int                 cursor_pos;
