@@ -8,7 +8,7 @@
  * Contributions from...
  * -----------------------------------------------
  * Peter Harvey		- pharvey@codebydesign.com
- * Nick Gorham      - nick@easysoft.com
+ * Nick Gorham      - nick@lurcher.org
  **************************************************/
 #include <config.h>
 #include <stdlib.h>
@@ -18,7 +18,9 @@ BOOL SQLGetConfigMode( UWORD *pnConfigMode )
 {
     inst_logClear();
 
+    __lock_config_mode();
     *pnConfigMode = __get_config_mode();
+    __unlock_config_mode();
 
     return TRUE;
 }
